class UserInterface:
    """An interface to Nion Swift UI.

    This class cannot be instantiated directly. Use :samp:`api_broker.get_api(version)` to get access an instance of
    this class.
    """

version = "~1.0"
