import numpy

class Interactive:
    @property
    def cancelled(self) -> bool:
        """Return True if the interactive script has been cancelled."""
        ...
    def print(self, text: str) -> None:
        """Print the text string."""
        ...
    def print_debug(self, text: str) -> None:
        """Print the text string if debug level output is enabled."""
        ...
    def print_info(self, text: str) -> None:
        """Print the text string if debug info output is enabled."""
        ...
    def print_warn(self, text: str) -> None:
        """Print the text string if warning level output is enabled."""
        ...
    def get_string(self, prompt: str, default_str: str=None) -> str:
        """Get a string from the user. Display the prompt. Optional default_str."""
        ...
    def get_integer(self, prompt: str, default_value: int=0) -> int: ...
    def get_float(self, prompt: str, default_value: float=0, format_str: str=None) -> float: ...
    def show_ndarray(self, data: numpy.ndarray, title:str = None) -> None: ...
    def confirm_ok_cancel(self, prompt: str) -> bool: ...
    def confirm_yes_no(self, prompt: str) -> bool: ...
    def confirm(self, prompt: str, accepted_text: str, rejected_text: str) -> bool: ...
    def alert(self, prompt: str, button_label: str = None) -> None: ...

version = "~1.0"
