"""
The MIME types used in the application.
"""

DISPLAY_ITEM_MIME_TYPE = "text/vnd.nionswift.display_item"
DISPLAY_PANEL_MIME_TYPE = "text/vnd.nionswift.display_panel"
DATA_SOURCE_MIME_TYPE = "text/vnd.nionswift.display_source"
GRAPHICS_MIME_TYPE = "text/vnd.nionswift.graphics"
DATA_GROUP_MIME_TYPE = "text/vnd.nionswift.data_group"
